/**====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*
 *
 * IOTCAPIs_WakeUp_Device.c
 *
 * Copyright (c) by TUTK Co.LTD. All Rights Reserved.
 *
 * @brief   Examle of how to use IOTC WakeUp API at Device side
 *
 *====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <errno.h>
#include "IOTCAPIs.h"
#include "IOTCWakeUp.h"

#define _KEEP_ALIVE_BY_TCP_ 1
// Macro definition
#define MAX_BUF_SIZE 1500

// Structure declaration
typedef struct _WakeupParam
{
    int                 nSkt;
    unsigned long       ulLoginAddr;
    unsigned short      usLoginPort;
    char                *pWakeupPattern;
    int 				nWakeupPatternLength;
} WakeupParam;

typedef struct{
	IOTC_WakeUpData* wakeupdata;
	int socketfd;
	char ip_port_str[128];
	
}TCPWakeupParam;



typedef struct _AliveParam
{
    int 			nSkt;
    IOTC_WakeUpData *pWUData;
    int 			nDataCnt;
} AliveParam;


// Variables
static bool g_bWakeUp = false;
IOTC_WakeUpData *gAliveData = NULL;
unsigned int gDataCnt = 0;

static int createSocket()
{
    int nSkt = -1;
#if _KEEP_ALIVE_BY_TCP_
	nSkt = socket(AF_INET, SOCK_STREAM, 0);
#else	
    nSkt = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
#endif
    printf("create socket = %d!\n", nSkt);

    return nSkt;
}


static void SetupSocketSelectParam(int nSkt, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *time, uint8_t nMaxTimeout)
{
    if(readfds)
    {
        FD_ZERO(readfds);
        FD_SET(nSkt, readfds);
    }

    if(writefds)
    {
        FD_ZERO(writefds);
        FD_SET(nSkt, writefds);
    }

    if(exceptfds)
    {
        FD_ZERO(exceptfds);
        FD_SET(nSkt, exceptfds);
    }

    if(time)
    {
        memset(time, 0, sizeof(struct timeval));
        time->tv_sec = nMaxTimeout;
    }
}


static bool isWakeupPattern(const char *pData, const char *pPattern, const int nPatternLength)
{
    bool bPattern = true;

    for(int i = 0; i < nPatternLength; i++)
    {
        if(pPattern[i])
        {
            if(pData[i] != pPattern[i])
            {
                bPattern = false;
                break;
            }
        }
    }

    return bPattern;
}


static void packetParser(WakeupParam *pParam)
{
    char buf[MAX_BUF_SIZE];
    int nRecvSize = 0;

    socklen_t socklength = 0;
    int nSkt = pParam->nSkt;

    //printf("[packetParser] - Prepare to receive socket header. skt = %d\n", nSkt);

    struct sockaddr_in remote;
    memset(&remote, 0, sizeof(struct sockaddr_in));
    remote.sin_family = AF_INET;
    remote.sin_addr.s_addr = pParam->ulLoginAddr;
    remote.sin_port = pParam->usLoginPort;

    nRecvSize = recvfrom(nSkt, &buf, MAX_BUF_SIZE, 0, (struct sockaddr *)&remote, &socklength);
    if (nRecvSize < 0)
    {
        printf("[packetParser] recvfrom error\n");
    }

    // Compare WakeUp pattern
    if(isWakeupPattern(buf, pParam->pWakeupPattern, pParam->nWakeupPatternLength))
        g_bWakeUp = true;
}


static void *thread_WakeUpTask(void *arg)
{
    fd_set readfds;
    struct timeval timeout;
    WakeupParam *pParam = (WakeupParam*)arg;

    int nSkt = pParam->nSkt;

    printf("[thread_WakeUpTask] - skt = %d\n", nSkt);

    int nVal = 0;
    int nMaxFds = nSkt + 1;
    while(1)
    {
        SetupSocketSelectParam(nSkt, &readfds, NULL, NULL, &timeout, 1);
        nVal = select(nMaxFds, &readfds, NULL, NULL, &timeout);
        if(nVal > 0)
        {
            if(FD_ISSET(nSkt, &readfds))
            {
                //printf("[thread_WakeUpTask] - FD_ISSET(pParam->nSkt, &readfds), skt = %d\n", nSkt);
                packetParser(pParam);
                //break;
            }
        } else if(nVal == 0) {
            //printf("[thread_WakeUpTask] - select timeout, skt = %d\n", nSkt);
        }

        if(g_bWakeUp)
        {
            printf("[thread_WakeUpTask] - Device is WakeUp!\n");
            break;
        }
    }

    printf("[testIOTC_Wakup] - thread_WakeUpTask exit!\n");
    pthread_exit(0);
}


static void *_thread_Login(void *arg)
{
    int nRet;
    char *UID = (char *)arg;

    while(1)
    {
        nRet = IOTC_Device_Login(UID, "AAAA0009", "12345678");
        printf("Calling IOTC_Device_Login() ret = %d\n", nRet);

        if(nRet == IOTC_ER_NoERROR)
            break;
        else
            sleep(2);
    }

    pthread_exit(0);
}


static void sendKeepAlivePacket(int nSkt, IOTC_WakeUpData *pIOTCAliveData, int nDataCnt)
{
    for(int i = 0; i < nDataCnt; i++)
    {
        struct sockaddr_in remote;
        memset(&remote, 0, sizeof(struct sockaddr_in));
        remote.sin_addr.s_addr = pIOTCAliveData[i].ulLoginAddr;
        remote.sin_port = pIOTCAliveData[i].usLoginPort;

        printf("[sendKeepAlivePacket] - %s : %d, size = %zu\n", inet_ntoa(remote.sin_addr), ntohs(remote.sin_port), pIOTCAliveData[i].nLoginPacketLength);

        int ret = sendto(nSkt, pIOTCAliveData[i].pszLoginPacket, pIOTCAliveData[i].nLoginPacketLength, 0, (struct sockaddr *)&remote, sizeof(struct sockaddr_in));
        if (ret < 0)
        {
            printf("[sendKeepAlivePacket] sendto error\n");
        }
    }
    
    //sleep(pIOTCAliveData[0].nLoginInterval);
    sleep(3);
}


static void duplicateWakeUpData(IOTC_WakeUpData *pDest, IOTC_WakeUpData* pSrc, int nDataCnt)
{
    if(pDest && pSrc)
    {
        for(int i = 0; i < nDataCnt; i++)
        {
            pDest[i].ulLoginAddr = pSrc[i].ulLoginAddr;   
            pDest[i].usLoginPort = pSrc[i].usLoginPort;
            pDest[i].nLoginInterval = pSrc[i].nLoginInterval;
            pDest[i].nLoginPacketLength = pSrc[i].nLoginPacketLength;
            pDest[i].nWakeupPatternLength = pSrc[i].nWakeupPatternLength;

            // copy Login packet
            pDest[i].pszLoginPacket = malloc(pSrc[i].nLoginPacketLength);
            memcpy(pDest[i].pszLoginPacket, pSrc[i].pszLoginPacket, pSrc[i].nLoginPacketLength);

            // copy wake up pattern
            pDest[i].pszWakeupPattern = malloc(pSrc[i].nWakeupPatternLength);
            memcpy(pDest[i].pszWakeupPattern, pSrc[i].pszWakeupPattern, pSrc[i].nWakeupPatternLength);
        }
    }
}


static void releaseWakeUpData(IOTC_WakeUpData *pData, int nDataCnt)
{
    if(pData)
    {
        for(int i = 0; i < nDataCnt; i++)
        {
            free(pData[i].pszLoginPacket);
            free(pData[i].pszWakeupPattern);
        }
        free(pData);
    }
}
#if _KEEP_ALIVE_BY_TCP_
static void *thread_WakeupByTCP(void *arg)
{
	printf("enter thread_WakeupByTCP...\n");
	TCPWakeupParam *tcp_wakeup_data = (TCPWakeupParam *)arg;
	if(!tcp_wakeup_data){
		printf("%s:arg invald!\n",__FUNCTION__);
		pthread_exit(-1);
	}

	while(1){
		if(g_bWakeUp){
			break;
		}
		char databuf[512] = {0};
		int ret = recv(tcp_wakeup_data->socketfd,databuf,512,0);
		if(ret > 0){
			printf("[%s]recv wakeup data:%d\n",tcp_wakeup_data->ip_port_str,ret);
			if(isWakeupPattern(databuf, tcp_wakeup_data->wakeupdata->pszWakeupPattern, tcp_wakeup_data->wakeupdata->nWakeupPatternLength)){
				printf("[%s]parse wakeup data result:true\n",tcp_wakeup_data->ip_port_str);
				//涉及到g_bWakeUp的地方，请自行加锁！
        		g_bWakeUp = true;
				break;
			}
		}
		else{
			printf("recv wakeup data error:%d\n",ret);
			break;
		}
	}
	close(tcp_wakeup_data->socketfd);	
	
	printf("leave thread_WakeupByTCP...\n");
	pthread_exit(0);
}

static void *thread_C2SandKeepAliveByTCP(void *arg)
{
	printf("enter thread_C2SandKeepAliveByTCP...\n");

	IOTC_WakeUpData *pParam = (IOTC_WakeUpData *)arg;
	
	if(!pParam){
		printf("%s:arg invald!\n",__FUNCTION__);
		pthread_exit(-1);
	}

	int   sockfd, n;
	struct sockaddr_in  servaddr;

	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		 printf("create socket error: %s(errno: %d)\n", strerror(errno),errno);
		 pthread_exit(-1);
	}
	
	//TCP连线P2P服务器
	memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(pParam->usLoginPort);
    servaddr.sin_addr.s_addr = pParam->ulLoginAddr;
	printf("[start] Connect Server->%s:%d\n",inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));
	
    if(connect(sockfd, (struct sockaddr*)&servaddr, sizeof(servaddr)) < 0){
        printf("connect error: %s(errno: %d)\n",strerror(errno),errno);
        pthread_exit(-1);
    }
	printf("[finished] Connect Server->%s:%d\n",inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));

	//创建wakeup线程
	TCPWakeupParam * tcp_wakeup_data = (TCPWakeupParam*)malloc(sizeof(TCPWakeupParam));
	tcp_wakeup_data->socketfd = sockfd;
	tcp_wakeup_data->wakeupdata = pParam;
	sprintf(tcp_wakeup_data->ip_port_str,"%s:%d",inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));
	pthread_t nTaskId;
	int nRet = pthread_create(&nTaskId, NULL, thread_WakeupByTCP, tcp_wakeup_data);
	
	//每隔nLoginInterval，对指定的某台服务器发送一次心跳包。TCP的情况下，可以略大于nLoginInterval，但是建议不超过1分钟。
	while(1){
		if(g_bWakeUp){
			break;
		}
		if(send(sockfd, pParam->pszLoginPacket, pParam->nLoginPacketLength, 0) < 0){
	        printf("send msg error: %s(errno: %d)\n", strerror(errno), errno);
	        break;
		}
		printf("Server->%s:%d, sockfd=%d send heartbeat successfully,sleep %d s\n",inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port),tcp_wakeup_data->socketfd,pParam->nLoginInterval);
		sleep(pParam->nLoginInterval);		
	}
	
	pthread_join(nTaskId,0);
		
	free(pParam->pszLoginPacket);
	free(pParam->pszWakeupPattern);
	free(pParam);
	free(tcp_wakeup_data);
	
	pParam->pszWakeupPattern = NULL;
	pParam->pszLoginPacket = NULL;
	pParam = NULL;
	tcp_wakeup_data = NULL;
	
	printf("leave thread_C2SandKeepAliveByTCP...\n");
	pthread_exit(0);
	
}
#endif


static void *thread_AliveTask(void *arg)
{
    AliveParam *ap = (AliveParam*)arg;	

    while(1)
    {
        if(!g_bWakeUp)
            sendKeepAlivePacket(ap->nSkt, ap->pWUData, ap->nDataCnt);
        else
            break;	
    }

    pthread_exit(0);
}


int IsDataReady(void)
{
    return gAliveData ? 1:0;
}

void SleepPacketCB(IOTC_WakeUpData *pData, unsigned int DataCnt, int nErrCode)
{
    printf("[%s] err [%d], cnt [%d]\n", __func__, nErrCode, DataCnt);
    gDataCnt = DataCnt;
    gAliveData = pData;
}

char UID[20] = {"C3PA8NJMU1VMUG6GUHZ1"};

int main(int argc, char *argv[])
{
    // ++++++ Devicde at normal mode ++++++
    if(argc >= 2)
    {
		strncpy(UID,argv[1],20);
    }

#if _KEEP_ALIVE_BY_TCP_
	printf("IOTCAPIs_Device[%s][TCP] start...\n",UID);
#else	
    printf("IOTCAPIs_Device[%s][UDP] start...\n",UID);
#endif
    int nRet;
    

    nRet = IOTC_Initialize2(0);

    printf("Step 1: IOTC_Initialize2() ret = %d\n", nRet);
    if(nRet != IOTC_ER_NoERROR)
    {
        printf("IOTCAPIs_Device exit...!!\n");
        return 0;
    }

    // Use Wake up feature
    IOTC_WakeUp_Init();
    // Login
    pthread_t threadID_Login;
    pthread_create(&threadID_Login, NULL, &_thread_Login, (void *)UID);
    pthread_detach(threadID_Login);

    sleep(3);


    //IOTC_WakeUp_Get_SleepPacket "Must" be used after login
    nRet = IOTC_WakeUp_Get_SleepPacket((IOTC_WakeUp_SleepPacketCB)SleepPacketCB, 0);
    if (nRet != IOTC_ER_NoERROR)
    {
        printf("IOTC_WakeUp_Get_SleepPacket error[%d]\n", nRet);
        return 0;
    }

    //wait until data is ready
    while(IsDataReady() == 0)
    {
        usleep(1000*10);
    }

    // Device is going to sleep, duplicate IOTC_WakeUpData
    IOTC_WakeUpData *pIOTCAliveData = NULL;
    pIOTCAliveData = (IOTC_WakeUpData*)malloc(sizeof(IOTC_WakeUpData) * gDataCnt);
    if(pIOTCAliveData)
        duplicateWakeUpData(pIOTCAliveData, gAliveData, gDataCnt);
    else
        printf("malloc error!\n");

    // Free gAliveData memory
    IOTC_WakeUp_DeInit(gAliveData);
    // DeInit IOTC
    IOTC_DeInitialize();
    // ------ Devicde at normal mode ------

#if _KEEP_ALIVE_BY_TCP_

	//此处可以对全部服务器做保活，也可以对部分服务器保活，建议对至少2台服务器做保活	
	for(int i = 0; i < gDataCnt; i++)
    {
    	pthread_t nTCPTaskId;
    	IOTC_WakeUpData *pParam = (IOTC_WakeUpData*)malloc(sizeof(IOTC_WakeUpData));
		
        pParam->ulLoginAddr = pIOTCAliveData[i].ulLoginAddr;
        pParam->usLoginPort = 8000;//此处可选用的端口：80/8000/8080/443/21047
        pParam->nWakeupPatternLength = pIOTCAliveData[i].nWakeupPatternLength;
        pParam->nLoginPacketLength = pIOTCAliveData[i].nLoginPacketLength;
		pParam->nLoginInterval = pIOTCAliveData[i].nLoginInterval;	
		
		pParam->pszWakeupPattern = (char*)malloc(pIOTCAliveData[i].nWakeupPatternLength+1);
		pParam->pszLoginPacket = (char*)malloc(pIOTCAliveData[i].nLoginPacketLength+1);
		
		memset(pParam->pszWakeupPattern,0,pIOTCAliveData[i].nWakeupPatternLength+1);
		memset(pParam->pszLoginPacket,0,pIOTCAliveData[i].nLoginPacketLength+1);
		
		memcpy(pParam->pszWakeupPattern,pIOTCAliveData[i].pszWakeupPattern,pParam->nWakeupPatternLength);
		memcpy(pParam->pszLoginPacket,pIOTCAliveData[i].pszLoginPacket,pParam->nLoginPacketLength);
	
        nRet = pthread_create(&nTCPTaskId, NULL, thread_C2SandKeepAliveByTCP,  pParam);

        if(nRet != 0)
            printf("pthread_create for fail!\n");
    }
	while(!g_bWakeUp){
		usleep(1000*10);
	}

#else
    // ++++++ Device enter sleep mode ++++++
    printf("Device enter sleep mode!\n");
    // Driver will create a receive thread and keep receive data
    int nSkt = createSocket();
	if(nSkt < 0){
		printf("createSocket failed,nSkt = %d\n",nSkt);
		return 1;
	}

    WakeupParam *pParam = (WakeupParam*)malloc(sizeof(WakeupParam) * gDataCnt);

    for(int i = 0; i < gDataCnt; i++)
    {
        pthread_t nTaskId;

        pParam[i].nSkt = nSkt;
        pParam[i].ulLoginAddr = pIOTCAliveData[i].ulLoginAddr;
        pParam[i].usLoginPort = pIOTCAliveData[i].usLoginPort;
        pParam[i].pWakeupPattern = pIOTCAliveData[i].pszWakeupPattern;
        pParam[i].nWakeupPatternLength = pIOTCAliveData[i].nWakeupPatternLength;

        nRet = pthread_create(&nTaskId, NULL, thread_WakeUpTask, &pParam[i]);

        if(nRet != 0)
            printf("pthread_create for fail!\n");
    }

    // Driver will create a send thread and keep send IOTC login
    pthread_t nAliveTaskId;

    AliveParam ap;
    ap.nSkt = nSkt;
    ap.pWUData = pIOTCAliveData;
    ap.nDataCnt = gDataCnt;

    nRet = pthread_create(&nAliveTaskId, NULL, thread_AliveTask, &ap);

    pthread_join(nAliveTaskId, 0);

    // Device is woken up
    close(nSkt);
    // Don't forget to free memory
    free(pParam);
#endif

    releaseWakeUpData(pIOTCAliveData, gDataCnt);
    // ------ Device enter sleep mode ------

    printf("Device enter normal mode!\n");

    nRet = IOTC_Initialize2(0);

    pthread_create(&threadID_Login, NULL, &_thread_Login, (void *)UID);
    pthread_detach(threadID_Login);
    //sleep(3);
    printf("Device prepare to listen!\n");

    int nSID = -1;
    while(1)
    {
        nSID = IOTC_Listen(0);
        if(nSID > -1){
			IOTC_Session_Close(nSID);
            break;
        }
        else
            sleep(1);
    }

    printf("Device accept connection\n");

    IOTC_DeInitialize();

    return 0;
}
